/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser;

import org.antlr.runtime.RecognitionException;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.GnuParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.backends.design.typecast.VerilogTypeCaster;
import ru.ispras.verilog.parser.backends.syntax.checker.VerilogStaticChecker;
import ru.ispras.verilog.parser.elaborator.VerilogDesign;
import ru.ispras.verilog.parser.elaborator.VerilogElaborator;
import ru.ispras.verilog.parser.model.VerilogNode;

/**
 * {@link VerilogTranslator} is the main class of the Verilog translator.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogTranslator {
  /**
   * {@link Parameters} implements the command line parameters parser.
   *
   * <p>It is based on the Apache's project {@code org.apache.commons.cli}.</p>
   */
  private static class Parameters {
    /** Command line parameter: -I, --include-path. */
    public static final String INCLUDE = "I";
    /** Command line parameter: -M, --module-name. */
    public static final String MODULE = "M";
    /** Command line parameter: -v, --verbose. */
    public static final String VERBOSE = "v";
    /** Command line parameter: -h, --help. */
    public static final String HELP = "h";

    /** Command line parameters. */
    private static Options options = new Options();

    static {
      options.addOption(INCLUDE, "include-path", true /* has value */,
          "Sets the include directories (a colon-separated list)");
      options.addOption(MODULE, "module-name", true /* has value */,
          "Sets the module to be elaborated");
      options.addOption(VERBOSE, "verbose", false /* has no value */,
          "Enables debug output");
      options.addOption(HELP, "help", false /* has no value */,
          "Shows this message");
    }

    /**
     * Constructs a parameters object.
     */
    private Parameters() {}

    /**
     * Parses the command line parameters.
     *
     * @param args the command line parameters.
     * @return the parsed structure of command line parameters.
     */
    public static CommandLine parse(final String[] args) throws ParseException {
      final CommandLineParser parser = new GnuParser();
      return parser.parse(options, args);
    }

    /**
     * Prints the help message.
     */
    public static void help() {
      final HelpFormatter formatter = new HelpFormatter();
      formatter.printHelp(80, "[options] verilog-files", "", options, "");
    }
  }

  /** Front-end. */
  private final VerilogFrontend frontend = new VerilogFrontend();
  /** AST-level back-ends. */
  private final VerilogSyntaxBackends syntaxBackends = new VerilogSyntaxBackends();
  /** Design-level back-ends. */
  private final VerilogDesignBackends designBackends = new VerilogDesignBackends();

  /** AST-based static checker. */
  private final VerilogSyntaxBackend checker =
      new VerilogSyntaxBackend("static-checker", new VerilogStaticChecker());

  /** Type caster for elaborated design. */
  private final VerilogDesignBackend caster = new VerilogTypeCaster("type-caster");

  /** Design elaboration engine. */
  private final VerilogElaborator elaborator =
      new VerilogElaborator("design-elaborator");

  public static void exit() {
    VerilogLogger.stat();
    InvariantChecks.checkFalse(VerilogLogger.failed());
  }

  /**
   * Constructs a Verilog translator.
   */
  public VerilogTranslator() {
    // Static checker goes first.
    add(checker);
    // Type casting goes after elaboration phase.
    add(caster);
  }

  /**
   * Registers an AST-level back-end in the processor.
   *
   * @param backend the back-end to be registered.
   */
  public final void add(final VerilogSyntaxBackend backend) {
    InvariantChecks.checkNotNull(backend);
    syntaxBackends.add(backend);
  }

  /**
   * Registers a design-level back-end in the processor.
   *
   * @param backend the back-end to be registered.
   */
  public final void add(final VerilogDesignBackend backend) {
    InvariantChecks.checkNotNull(backend);
    designBackends.add(backend);
  }

  /**
   * Launches the Verilog translator.
   *
   * @param args the command line parameters.
   */
  public final void start(final String[] args) {
    VerilogLogger.reset();

    try {
      final CommandLine params = Parameters.parse(args);

      // Handle the --verbose parameter.
      VerilogLogger.verbose(params.hasOption(Parameters.VERBOSE));

      // Handle the --include-path parameter.
      if (params.hasOption(Parameters.INCLUDE)) {
        VerilogLogger.debug("Include path: %s", params.getOptionValue(Parameters.INCLUDE));
        frontend.addPath(params.getOptionValue(Parameters.INCLUDE));
      }

      // Handle the --module-name parameter.
      if (params.hasOption(Parameters.MODULE)) {
        VerilogLogger.debug("Module name: %s", params.getOptionValue(Parameters.MODULE));
        elaborator.setModuleName(params.getOptionValue(Parameters.MODULE));

        // Add the elaboration back-end if the module name has been set.
        add(elaborator);
      }

      // Handle the --help parameter.
      if (params.hasOption(Parameters.HELP)) {
        Parameters.help();
      }

      try {
        // Parse the code and construct the abstract syntax tree.
        final VerilogNode root = frontend.start(params.getArgs());

        // Append the standard declarations (system tasks and functions, primitive gates, etc.).
        root.addAll(VerilogLibrary.get().getDeclarations());

        // Launch syntax-based back-ends.
        syntaxBackends.start(root);

        // Launch design-based back-ends.
        if (params.hasOption(Parameters.MODULE)) {
          final VerilogDesign design = elaborator.getDesign();
          InvariantChecks.checkNotNull(design);

          designBackends.start(design);
        }
      } catch (final RecognitionException e) {
        // Source code parsing exception.
        VerilogLogger.error("[Parsing] %s", e.getMessage());
        e.printStackTrace();
      }
    } catch (final ParseException e) {
      // Command line parsing exception.
      Parameters.help();
    } catch (final Throwable e) {
      VerilogLogger.error("[Internal] %s", e.getMessage());
      e.printStackTrace();
    }

    VerilogTranslator.exit();
  }
}
