/*
 * Copyright 2018-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.elaborator;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.Variable;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.transformer.NodeProvider;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.transformer.ValueProvider;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.fortress.util.Pair;
import ru.ispras.verilog.parser.VerilogLogger;
import ru.ispras.verilog.parser.VerilogSyntaxBackend;
import ru.ispras.verilog.parser.VerilogTranslator;
import ru.ispras.verilog.parser.backends.syntax.checker.VerilogStaticChecker;
import ru.ispras.verilog.parser.core.AbstractNode;
import ru.ispras.verilog.parser.interpreter.VerilogCalculator;
import ru.ispras.verilog.parser.model.VerilogActivity;
import ru.ispras.verilog.parser.model.VerilogAssign;
import ru.ispras.verilog.parser.model.VerilogAssignment;
import ru.ispras.verilog.parser.model.VerilogCaseGenerate;
import ru.ispras.verilog.parser.model.VerilogCaseGenerateItem;
import ru.ispras.verilog.parser.model.VerilogDeclaration;
import ru.ispras.verilog.parser.model.VerilogDefineParameter;
import ru.ispras.verilog.parser.model.VerilogGenerate;
import ru.ispras.verilog.parser.model.VerilogIfGenerate;
import ru.ispras.verilog.parser.model.VerilogInstantiation;
import ru.ispras.verilog.parser.model.VerilogLoopGenerate;
import ru.ispras.verilog.parser.model.VerilogModule;
import ru.ispras.verilog.parser.model.VerilogNode;
import ru.ispras.verilog.parser.model.VerilogNode.Tag;
import ru.ispras.verilog.parser.model.VerilogPort;
import ru.ispras.verilog.parser.model.VerilogPortConnection;
import ru.ispras.verilog.parser.model.VerilogProcedure;
import ru.ispras.verilog.parser.model.VerilogStatement;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;
import ru.ispras.verilog.parser.model.basis.VerilogMinTypMax;
import ru.ispras.verilog.parser.model.basis.VerilogPath;
import ru.ispras.verilog.parser.model.basis.VerilogPathItem;
import ru.ispras.verilog.parser.model.basis.VerilogReference;
import ru.ispras.verilog.parser.model.util.ModelUtils;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * {@link VerilogElaborator} implements a simple elaboration engine.
 *
 * <p>
 * [IEEE Std 1364-2005]
 *
 * 12.8 Elaboration
 *
 * Elaboration is the process that occurs between parsing and simulation. It binds modules to
 * module instances, builds the model hierarchy, computes parameter values, resolves hierarchical
 * names, establishes net connectivity, and prepares all of this for simulation. With the addition
 * of generate constructs, the order in which these tasks occur becomes significant.
 * </p>
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogElaborator extends VerilogSyntaxBackend {
  /** Simple solution against recursion. */
  private static final int STARTING_POINTS_LIMIT = 512;

  /** Temporal variable number. */
  private static int variableNumber = 0;

  /**
   * Returns a unique identifier to be used as a variable name.
   *
   * @param prefix the prefix.
   * @return a unique identifier.
   */
  static String newVariableName(final String prefix) {
    return String.format("%s%d", (prefix.isEmpty() ? "tmp" : prefix), variableNumber++);
  }

  /** Elaborated design builder. */
  private final VerilogDesign.Builder designBuilder = new VerilogDesign.Builder();

  /** Name of the module to be elaborated. */
  private String moduleName;

  /** Elaborated design. */
  private VerilogDesign design;

  /**
   * Creates an elaboration engine.
   *
   * @param name the back-end name.
   */
  public VerilogElaborator(final String name) {
    super(name);
    VerilogElaborator.variableNumber = 0;
  }

  /**
   * Creates an elaboration engine.
   *
   * @param name the back-end name.
   * @param moduleName the name of the module to be elaborated.
   */
  public VerilogElaborator(final String name, final String moduleName) {
    this(name);

    InvariantChecks.checkNotNull(moduleName);
    this.moduleName = moduleName;
  }

  /**
   * Returns the name of the module to be elaborated.
   *
   * @return the module name.
   */
  public String getModuleName() {
    return moduleName;
  }

  /**
   * Sets the name of the module to be elaborated.
   *
   * @param moduleName the module name.
   */
  public void setModuleName(final String moduleName) {
    InvariantChecks.checkNotNull(moduleName);
    this.moduleName = moduleName;
  }

  /**
   * Returns the elaborated design.
   *
   * @return the elaborated design.
   */
  public VerilogDesign getDesign() {
    return design;
  }

  @Override
  public void start(final VerilogNode rootNode) {
    InvariantChecks.checkNotNull(rootNode);

    /*
     * [IEEE Std 1364-2005]
     *
     * 12.8.1 Order of elaboration
     *
     * Because of generate constructs, the model hierarchy can depend on parameter values. Because
     * defparam statements can alter parameter values from almost anywhere in the hierarchy, the
     * result of elaboration can be ambiguous when generate constructs are involved. The final model
     * hierarchy can depend on the order in which defparams and generate constructs are evaluated.
     *
     * The following algorithm defines an order that produces the correct hierarchy:
     *
     * a) A list of starting points is initialized with the list of top-level modules.
     *
     * b) The hierarchy below each starting point is expanded as much as possible without
     *    elaborating generate constructs. All parameters encountered during this expansion are
     *    given their final values by applying initial values, parameter overrides, and defparam
     *    statements.
     *
     *    In other words, any defparam statement whose target can be resolved within the hierarchy
     *    elaborated so far must have its target resolved and its value applied. defparam statements
     *    whose target cannot be resolved are deferred until the next iteration of this step.
     *    Because no defparam inside the hierarchy below a generate construct is allowed to refer to
     *    a parameter outside the generate construct, it is possible for parameters to get their
     *    final values before going to step c).
     *
     * c) Each generate construct encountered in step b) is revisited, and the generate scheme is
     *    evaluated. The resulting generate block instantiations make up the new list of starting
     *    points. If the new list of starting points is not empty, go to step b).
     */

    // Step a): compose the list of starting points.
    List<Pair<VerilogNode, VerilogContext>> startingPoints = getStartingPoints(rootNode);

    if (!startingPoints.isEmpty()) {
      // The elaborated design does not keep the hierarchy.
      designBuilder.setName(startingPoints.iterator().next().first.getName());
    }

    while (!startingPoints.isEmpty()) {
      // The points to be handled by the generate processor.
      List<Pair<VerilogNode, VerilogContext>> generatePoints = new ArrayList<>(startingPoints);

      // The hierarchy is expanded as much as possible without elaborating generate constructs.
      while (!startingPoints.isEmpty()) {
        // Temporary solution against recursion.
        if (generatePoints.size() > STARTING_POINTS_LIMIT) {
          VerilogLogger.error("Starting points limit has been exhausted: %", generatePoints.size());
          VerilogTranslator.exit();
        }

        final List<Pair<VerilogNode, VerilogContext>> newStartingPoints = new ArrayList<>();

        // Step b): expand each starting point.
        for (final Pair<VerilogNode, VerilogContext> startingPoint : startingPoints) {
          newStartingPoints.addAll(
            expand(
              startingPoint.first  /* Elaborated node */,
              startingPoint.second /* Elaboration context */
            )
          );
        }

        startingPoints = newStartingPoints;
        generatePoints.addAll(newStartingPoints);
      }

      final List<Pair<VerilogNode, VerilogContext>> newStartingPoints = new ArrayList<>();

      // Step c): expand each starting point.
      for (final Pair<VerilogNode, VerilogContext> startingPoint : generatePoints) {
        newStartingPoints.addAll(
          generate(
            startingPoint.first  /* Elaborated node */,
            startingPoint.second /* Elaboration context */
          )
        );
      }

      startingPoints = newStartingPoints;
    }

    // Finalize the elaboration process.
    design = designBuilder.build();
  }

  /**
   * Performs step a) of the elaboration process.
   *
   * <p>
   * [IEEE Std 1364-2005]
   *
   * a) A list of starting points is initialized with the list of top-level modules.
   * </p>
   *
   * @param rootNode the AST root
   * @return the list of starting points
   */
  private List<Pair<VerilogNode, VerilogContext>> getStartingPoints(final VerilogNode rootNode) {
    InvariantChecks.checkNotNull(rootNode);

    final List<Pair<VerilogNode, VerilogContext>> startingPoints = new ArrayList<>();

    // NOTE: There is one starting point, i.e. only one top-level module is elaborated.
    final AbstractNode<Tag> node = rootNode.find(moduleName);

    if (node != null && node.getTag() == VerilogModule.TAG) {
      final VerilogNode module = (VerilogNode) node;
      final VerilogContext context = new VerilogContext();

      context.begin(module);
      startingPoints.add(new Pair<>(module, context.clone()));
      context.end();
    } else {
      VerilogLogger.error("Module '%s' has not been found", moduleName);
    }

    return startingPoints;
  }

  /**
   * Performs step b) of the elaboration process.
   *
   * <p>
   * [IEEE Std 1364-2005]
   *
   * b) The hierarchy below each starting point is expanded as much as possible without
   *    elaborating generate constructs. All parameters encountered during this expansion are
   *    given their final values by applying initial values, parameter overrides, and defparam
   *    statements.
   *
   *    In other words, any defparam statement whose target can be resolved within the hierarchy
   *    elaborated so far must have its target resolved and its value applied. defparam statements
   *    whose target cannot be resolved are deferred until the next iteration of this step.
   *    Because no defparam inside the hierarchy below a generate construct is allowed to refer to
   *    a parameter outside the generate construct, it is possible for parameters to get their
   *    final values before going to step c).
   * </p>
   *
   * @param node the node to be expanded.
   * @param context the elaboration context.
   * @return the list of new starting points.
   */
  private List<Pair<VerilogNode, VerilogContext>> expand(
      final VerilogNode node,
      final VerilogContext context) {
    InvariantChecks.checkNotNull(node);
    InvariantChecks.checkNotNull(context);

    VerilogLogger.debug("Expanding node '%s'...", node);
    final List<Pair<VerilogNode, VerilogContext>> newStartingPoints = new ArrayList<>();

    // Create bindings for the instantiation.
    final VerilogNode instanceNode = context.getNode();

    createVariablesAndBindings(node, context);
    VerilogLogger.debug("Bindings: %s", context.getBindings());

    // Handle the parameters' overrides.
    if (instanceNode != null) {
      final Collection<AbstractNode<Tag>> parameters = instanceNode.items(Tag.ASSIGNMENT);
      InvariantChecks.checkNotNull(parameters);

      for (final AbstractNode<Tag> parameter : parameters) {
        final VerilogAssignment assignmentNode = (VerilogAssignment) parameter;
        defineParameter(assignmentNode, context);
      }
    }

    // Instantiate the module's processes and continuous assignments.
    final Collection<AbstractNode<Tag>> processes = node.items(Tag.ACTIVITY, Tag.ASSIGN, Tag.DECLARATION);
    InvariantChecks.checkNotNull(processes);

    for (final AbstractNode<Tag> process : processes) {
      /*
       * [IEEE Std 1364-2005]
       * 
       * 6.1.1 The net declaration assignment
       * 
       * The net declaration assignment allows a continuous assignment to be placed
       * on a net in the same statement that declares the net.
       */
      if (process.getTag() == Tag.DECLARATION) {
        final VerilogDeclaration declaration = (VerilogDeclaration) process;

        if (!declaration.getElementType().isWire() || declaration.getExpression() == null) {
          continue;
        }
      }

      final VerilogNode processNode = (VerilogNode) process;
      addProcess(processNode, context, newStartingPoints);
    }

    // Handle the define parameter statements.
    final Collection<AbstractNode<Tag>> defines = node.items(Tag.DEFINE_PARAMETER);
    InvariantChecks.checkNotNull(defines);

    for (final AbstractNode<Tag> define : defines) {
      final VerilogDefineParameter defineNode = (VerilogDefineParameter) define;
      defineParameter(defineNode, context);
    }

    // Handle the inner instantiations (except for generate-wrapped ones).
    final Collection<AbstractNode<Tag>> subinstances = node.items(Tag.INSTANTIATION);
    InvariantChecks.checkNotNull(subinstances);

    for (final AbstractNode<Tag> subinstance : subinstances) {
      final VerilogInstantiation subinstanceNode = (VerilogInstantiation) subinstance;
      final VerilogModule submoduleNode = subinstanceNode.getDeclaration();

      if (submoduleNode != null) {
        if (submoduleNode.isModule()) {
          context.begin(subinstanceNode);
          newStartingPoints.add(new Pair<>(submoduleNode, context.clone()));
          context.end();
        } else {
          // TODO: Support gates, switches, and user-defined primitives.
        }
      } else {
        VerilogLogger.error("Module '%s' has not been found", subinstanceNode.getModuleName());
        VerilogTranslator.exit();
      }
    }

    return newStartingPoints;
  }

  /**
   * Performs step b) of the elaboration process.
   *
   * <p>
   * c) Each generate construct encountered in step b) is revisited, and the generate scheme is
   *    evaluated. The resulting generate block instantiations make up the new list of starting
   *    points. If the new list of starting points is not empty, go to step b).
   * </p>
   *
   * @param node the node to be generated.
   * @param context the elaboration context.
   * @return the list of new starting points.
   */
  private List<Pair<VerilogNode, VerilogContext>> generate(
      final VerilogNode node,
      final VerilogContext context) {
    InvariantChecks.checkNotNull(node);
    InvariantChecks.checkNotNull(context);

    VerilogLogger.debug("Generating module '%s'...", node);

    final Map<String, VerilogParameter> parameters = designBuilder.getParameters();
    final List<Pair<VerilogNode, VerilogContext>> newStartingPoints = new ArrayList<>();

    // Handle generate constructs.
    final Collection<AbstractNode<Tag>> generates = node.items(Tag.GENERATE);
    InvariantChecks.checkNotNull(generates);

    for (final AbstractNode<Tag> generate : generates) {
      final VerilogGenerate generateNode = (VerilogGenerate) generate;
      InvariantChecks.checkNotNull(generateNode);

      newStartingPoints.add(new Pair<>(generateNode, context.clone()));
    }

    // Handle if-then-else generate constructs.
    final Collection<AbstractNode<Tag>> generateIfThens = node.items(Tag.IF_GENERATE);
    InvariantChecks.checkNotNull(generateIfThens);

    for (final AbstractNode<Tag> generateIfThen : generateIfThens) {
      final VerilogIfGenerate ifGenerate = (VerilogIfGenerate) generateIfThen;

      final Node oldExpression = ifGenerate.getExpression().getNode();
      final Node newExpression = substitute(oldExpression, context.getBindings());

      if (evaluateBoolean(newExpression, context, parameters)) {
        final VerilogNode generateNode = ifGenerate.getThenGenerate();
        InvariantChecks.checkNotNull(generateNode);

        newStartingPoints.add(new Pair<>(generateNode, context.clone()));
      } else {
        final VerilogNode generateNode = ifGenerate.getElseGenerate();

        // The else branch may be missing.
        if (generateNode != null) {
         newStartingPoints.add(new Pair<>(generateNode, context.clone()));
        }
      }
    }

    // Handle case generate constructs.
    final Collection<AbstractNode<Tag>> generateCases = node.items(Tag.CASE_GENERATE);
    InvariantChecks.checkNotNull(generateCases);

    for (final AbstractNode<Tag> generateCase : generateCases) {
      final VerilogCaseGenerate caseGenerate = (VerilogCaseGenerate) generateCase;

      for (final VerilogCaseGenerateItem caseGenerateItem : caseGenerate.getCases()) {
        final Node oldExpression = caseGenerateItem.getExpression().getNode();
        final Node newExpression = substitute(oldExpression, context.getBindings());

        if (evaluateBoolean(newExpression, context, parameters)) {
          final VerilogNode generateNode = caseGenerateItem.getGenerate();
          InvariantChecks.checkNotNull(generateNode);

          newStartingPoints.add(new Pair<>(generateNode, context.clone()));
          break;
        }
      }
    }

    // Handle for generate constructs.
    final Collection<AbstractNode<Tag>> generateLoops = node.items(Tag.LOOP_GENERATE);
    InvariantChecks.checkNotNull(generateLoops);

    for (final AbstractNode<Tag> generateLoop : generateLoops) {
      final VerilogLoopGenerate loopGenerate = (VerilogLoopGenerate) generateLoop;

      final Node oldExpression = loopGenerate.getExpression().getNode();
      final Node newExpression = substitute(oldExpression, context.getBindings());

      final VerilogAssignment initialization = loopGenerate.getInitialization();
      final VerilogAssignment iteration = loopGenerate.getIteration();

      defineGenvar(initialization, context);

      while (true) {
        if (!evaluateBoolean(newExpression, context, parameters)) {
          break;
        }

        newStartingPoints.add(new Pair<>(loopGenerate.getGenerate(), context.clone()));
        defineGenvar(iteration, context);
      }
    }

    return newStartingPoints;
  }

  private void createVariableAndBinding(
      final VerilogVariable.Type variableType,
      final NodeVariable oldVariable,
      final NodeVariable newVariable,
      final VerilogBinding.Type bindingType,
      final Node rhs,
      final VerilogDeclaration declaration,
      final VerilogContext context) {
    final VerilogVariable lhs = new VerilogVariable(variableType, oldVariable, newVariable);
    final VerilogBinding binding = new VerilogBinding(bindingType, lhs, rhs);
System.out.println("VAR: " + variableType + ", " + oldVariable.getName() + ", " + newVariable.getName() + ", " + bindingType + ", " + rhs );
    context.getBindings().put(oldVariable.getName(), binding);

    if (declaration.isParam()) {
      defineParameter(declaration, context);
    } else {
      final VerilogInstantiator instantiator = new VerilogInstantiator(declaration, context, false);
      designBuilder.addVariable(lhs, instantiator);
    }
  }

  private void addProcess(
      final VerilogNode node,
      final VerilogContext context,
      final List<Pair<VerilogNode, VerilogContext>> newStartingPoints) {
    InvariantChecks.checkTrue(node.isActivity() || node.isAssign() || node.isDeclaration());

    final VerilogNode copy = node.clone();
    VerilogInstantiator instantiator;

    if (node.isActivity() || node.isAssign()) {
      // NOTE: The call collector modifies the node if it contains function calls.
      // It is not a problem that the collector modifies the node (replaces calls with variables);
      // the problem is that further instantiation of the process will not instantiate functions.

      // TODO: Do we really need a copy?
      final Collection<Pair<VerilogDeclaration, VerilogExpression>> functionCalls =
          VerilogCallCollector.transform(copy);

      for (final Pair<VerilogDeclaration, VerilogExpression> functionCall : functionCalls) {
        addFunctionCall(functionCall.first, functionCall.second, context, newStartingPoints);
      }

      instantiator = copy.isActivity()
          ? new VerilogInstantiator((VerilogActivity) copy, context.clone())
          : new VerilogInstantiator((VerilogAssign) copy, context.clone());
    } else {
      // Declaration case.
      final VerilogDeclaration declaration = (VerilogDeclaration) copy;
      instantiator = new VerilogInstantiator(declaration, context.clone(), true);
    }

    designBuilder.addProcess(instantiator);
  }

  private void addFunctionCall(
      final VerilogDeclaration returnDeclaration,
      final VerilogExpression functionCall,
      final VerilogContext context,
      final List<Pair<VerilogNode, VerilogContext>> newStartingPoints) {
    final VerilogProcedure functionNode = functionCall.getDeclaration();

    if (functionNode == null) {
      VerilogLogger.error("Function '%s' has not been found", functionCall.getPath());
      VerilogTranslator.exit();
    }

    VerilogLogger.debug("Declaration of '%s' in %s has been found: %s",
        functionCall.getPath(), functionCall.getNode(), functionNode);

    // Convert the function declaration to the module declaration (reuse).
    final VerilogModule moduleNode = new VerilogModule(functionNode.getParentNode());
    moduleNode.setName(functionNode.getName());

    final VerilogActivity processNode = new VerilogActivity(moduleNode);
    processNode.setType(VerilogActivity.Type.ALWAYS);

    final VerilogStatement statementNode = functionNode.getStatement();
    final VerilogStatement statementCopy = statementNode.clone();

    statementCopy.setParent(processNode);
    processNode.setStatement(statementCopy);

    moduleNode.add(processNode);

    // Convert the function call to the module instantiation (reuse).
    final VerilogInstantiation instanceNode = new VerilogInstantiation(functionNode.getParentNode());
    instanceNode.setName(String.format("%s_instance", returnDeclaration.getName()));
    instanceNode.setModuleName(functionNode.getName());

    final Collection<AbstractNode<Tag>> declarations = functionNode.items(Tag.DECLARATION);
    InvariantChecks.checkNotNull(declarations);

    final NodeOperation operation = functionCall.getOperation();
    // Start with the operand #1 (operand #0 stores the function declaration).
    int operandIndex = 1;

    for (final AbstractNode<Tag> declaration : declarations) {
      final VerilogDeclaration declarationNode = (VerilogDeclaration) declaration;
      final VerilogDeclaration declarationCopy = declarationNode.clone();

      declarationCopy.setParent(moduleNode);
      moduleNode.addDeclaration(declarationCopy);

      if (declarationCopy.isPort()) {
        final VerilogPortConnection connection = new VerilogPortConnection(instanceNode);

        connection.setName(declarationCopy.getName());

        if (declarationCopy.isInput()) {
          // Bind the argument.
          final Node operand = operation.getOperand(operandIndex++);
          final VerilogExpression argumentExpression = new VerilogExpression(operand);

          connection.setExpression(argumentExpression);
        } else {
          // Add the return-value variable to the design.
          createVariablesAndBindings(returnDeclaration, context);

          // Bind the return-value variable to the module's output.
          final VerilogPathItem returnItem = new VerilogPathItem(returnDeclaration.getName());

          final VerilogPath returnPath = new VerilogPath();
          returnPath.addItem(returnItem);

          final VerilogReference returnReference = new VerilogReference();
          returnReference.setPath(returnPath);
          returnReference.setDeclaration(returnDeclaration);

          final VerilogExpression returnExpression = new VerilogExpression(returnReference.getNode());

          connection.setExpression(returnExpression);
        }

        instanceNode.addConnection(connection);
      }
    }

    // TODO: Get back-ends from the translator.
    final VerilogSyntaxBackend checker = new VerilogSyntaxBackend("static-checker", new VerilogStaticChecker());
    checker.start(moduleNode);
    checker.start(instanceNode);

    context.begin(instanceNode, true /* Reuse parent bindings */);
    newStartingPoints.add(new Pair<>(moduleNode, context.clone()));
    context.end();
  }

  private void createVariablesAndBindings(
      final VerilogNode node,
      final VerilogContext context) {
    final Set<String> elaboratedNames = new HashSet<>();

    if (node.isModule()) {
      final VerilogModule module = (VerilogModule) node;

      // Create bindings for the module's ports.
      final Collection<VerilogPort> ports = module.getPorts();
      for (final VerilogPort port : ports) {
        createVariablesAndBindings(port, context);

        final Collection<VerilogReference> references = port.getReferences();
        for (final VerilogReference reference : references) {
          // Add port references to the elaborated names.
          elaboratedNames.add(reference.getPath().toString());
        }
      }
    }

    // Create bindings for the module's declarations.
    final Collection<AbstractNode<Tag>> declarations = node.items(Tag.DECLARATION);
    for (final AbstractNode<Tag> declaration : declarations) {
      final VerilogDeclaration declarationNode = (VerilogDeclaration) declaration;

      // Except for the ports.
      if (!elaboratedNames.contains(declarationNode.getName())) {
        createVariablesAndBindings(declarationNode, context);
        elaboratedNames.add(declarationNode.getName());
      }
    }
  }

  private void createVariablesAndBindings(
      final VerilogPort portNode,
      final VerilogContext context) {
    // Bindings for the references.
    final Collection<VerilogReference> references = portNode.getReferences();
    for (final VerilogReference referenceNode : references) {
      createVariablesAndBindings(referenceNode, context);
    }
  }

  private void createVariablesAndBindings(
      final VerilogDeclaration declarationNode,
      final VerilogContext context) {
    if (declarationNode.isGenvar()) {
      defineGenvar(declarationNode, context);
    } else {
      // Instance variables and parameters have the global names.
      final NodeVariable oldVariable = ModelUtils.getVariable(declarationNode);
      final NodeVariable newVariable = context.getVariable(oldVariable);

      // Binding is required for variables and parameters.
      createVariableAndBinding(
          VerilogVariable.Type.VARIABLE,
          oldVariable,
          newVariable,
          VerilogBinding.Type.FINAL,
          newVariable,
          declarationNode,
          context
      );
    }
  }

  private void createVariablesAndBindings(
      final VerilogReference referenceNode,
      final VerilogContext context) {
    final VerilogDeclaration declarationNode = referenceNode.getDeclaration();
    final NodeVariable oldVariable = ModelUtils.getVariable(referenceNode);
    final NodeVariable newVariable = context.getVariable(oldVariable);
    final VerilogNode node = context.getNode();

    final Collection<AbstractNode<Tag>> portConnections = node.items(Tag.PORT_CONNECTION);

    // A top-level module has no port connections (node is module);
    // an inner module has port connections: .port(expression).
    Node oldExpression = null;
    for (final AbstractNode<Tag> portConnection : portConnections) {
      final VerilogPortConnection portConnectionNode = (VerilogPortConnection) portConnection;
      final String portName = portConnectionNode.getName();
      InvariantChecks.checkNotNull(portName);

      if (portName.equals(oldVariable.getName())) {
        final VerilogExpression portExpression = portConnectionNode.getExpression();

        // It is possible that port expression is absent, i.e. null.
        oldExpression = portExpression != null ? portExpression.getNode() : null;
        break;
      }
    }

    // Port expression is absent.
    if (oldExpression == null) {
      createVariableAndBinding(
          VerilogVariable.Type.VARIABLE,
          oldVariable,
          newVariable,
          VerilogBinding.Type.FINAL,
          newVariable,
          declarationNode,
          context
      );
      return;
    }

    // [IEEE Std 1364-2005]
    //
    // 12.3.9.2 Rule 2
    //
    // Each port connection shall be a continuous assignment of source to sink, where one connected
    // item shall be a signal source and the other shall be a signal sink. The assignment shall be
    // a continuous assignment from source to sink for input or output ports. The assignment is a
    // nonstrength reducing transistor connection for inout ports. Only nets or structural net
    // expressions shall be the sinks in an assignment.

    /*
    final VerilogReference lhs = new VerilogReference();
    final VerilogExpression rhs = new VerilogExpression();

    if (declarationNode.isInput()) {
      lhs.setNode(newVariable); // Global name.
      rhs.setNode(oldExpression);
    } else if (declarationNode.isOutput()) {
      lhs.setNode(oldExpression);
      rhs.setNode(newVariable); // Global name.
    } else {
      VerilogLogger.error("Instantiations with inout ports are unsupported");
      VerilogTranslator.exit();
    }

    final VerilogAssign assignNode = new VerilogAssign(node);
    final VerilogAssignment assignmentNode = new VerilogAssignment(assignNode);
    assignmentNode.addReference(lhs);
    assignmentNode.setRhsExpression(new VerilogMinTypMax(rhs));
    assignNode.setAssignment(assignmentNode);

    final VerilogContext parentContext = context.clone();
    context.end();

    parentContext.getBindings().put(newVariable.getName(), new VerilogBinding(new VerilogVariable(newVariable, newVariable), newVariable));
    addProcess(assignNode, parentContext, null);
    */

    // Port connections of inner module instances are mapped to upper-level expressions.
    createVariableAndBinding(
        VerilogVariable.Type.BINDING,
        oldVariable,
        newVariable,
        VerilogBinding.Type.PARENT_CONTEXT,
        oldExpression,
        declarationNode,
        context
    );
  }

  private void defineGenvar(
      final VerilogDeclaration declarationNode,
      final VerilogContext context) {
    InvariantChecks.checkNotNull(declarationNode);
    InvariantChecks.checkNotNull(context);

    // Generate variable's name is not transformed to the full one.
    final String genvarName = declarationNode.getName();

    final VerilogMinTypMax oldMinTypMax = declarationNode.getExpression();

    if (oldMinTypMax != null) {
      final Node oldExpression = oldMinTypMax.getMinExpression().getNode();
      final Node newExpression = substitute(oldExpression, context.getBindings());

      final Map<String, VerilogParameter> parameters = designBuilder.getParameters();
      final Data data = evaluate(newExpression, context, parameters);

      context.addGenvar(genvarName, data);
    } else {
      // Define the variable with the default value.
      context.addGenvar(genvarName, Data.newInteger(0));
    }
  }

  private void defineGenvar(
      final VerilogAssignment assignmentNode,
      final VerilogContext context) {
    InvariantChecks.checkNotNull(assignmentNode);
    InvariantChecks.checkNotNull(context);

    final VerilogReference reference = assignmentNode.getReferences().iterator().next();
    final VerilogExpression expression = assignmentNode.getRhsExpression().getMinExpression();

    // Generate variable's name is not transformed to the full one.
    final String genvarName = reference.getPath().toString();

    final Node oldExpression = expression.getNode();
    final Node newExpression = substitute(oldExpression, context.getBindings());

    final Map<String, VerilogParameter> parameters = designBuilder.getParameters();
    final Data data = evaluate(newExpression, context, parameters);

    context.addGenvar(genvarName, data);
  }

  private void defineParameter(
      final VerilogDeclaration declarationNode,
      final VerilogContext context) {
    InvariantChecks.checkNotNull(declarationNode);
    InvariantChecks.checkNotNull(context);

    // Parameter's name is transformed to the full one.
    final String newParameterName = context.getFullName(declarationNode.getName());

    final Node oldExpression = declarationNode.getExpression().getMinExpression().getNode();
    final Node newExpression = substitute(oldExpression, context.getBindings());

    final Map<String, VerilogParameter> parameters = designBuilder.getParameters();
    final Data data = evaluate(newExpression, context, parameters);
    InvariantChecks.checkNotNull(data, String.format("expression=%s", newExpression));

    VerilogLogger.debug("Declare parameter: %s = %s = %s", newParameterName, newExpression, data);

    switch (declarationNode.getType()) {
      case PARAMETER:
        designBuilder.addParameter(VerilogParameter.Type.PARAMETER, newParameterName, data);
        break;
      case SPECPARAM:
        designBuilder.addParameter(VerilogParameter.Type.SPECPARAM, newParameterName, data);
        break;
      case LOCALPARAM:
        designBuilder.addParameter(VerilogParameter.Type.LOCALPARAM, newParameterName, data);
        break;
      default:
        break;
    }
  }

  private void defineParameter(
      final VerilogAssignment assignmentNode,
      final VerilogContext context) {
    InvariantChecks.checkNotNull(assignmentNode);

    final VerilogReference reference = assignmentNode.getReferences().iterator().next();
    final VerilogMinTypMax expression = assignmentNode.getRhsExpression();

    // NOTE: Expression is allowed to be missing.
    if (expression != null) {
      final String newParameterName = context.getFullName(reference.getPath().toString());
      final Map<String, VerilogBinding> parentBindings = context.getParentBindings();

      final Node oldExpression = expression.getMinExpression().getNode();
      final Node newExpression = substitute(oldExpression, parentBindings);

      final Map<String, VerilogParameter> parameters = designBuilder.getParameters();
      final Data data = evaluate(newExpression, context, parameters);
      InvariantChecks.checkNotNull(data, String.format("expression=%s", newExpression));

      VerilogLogger.debug("Override parameter: %s = %s = %s", newParameterName, newExpression, data);
      designBuilder.addParameter(VerilogParameter.Type.OVERRIDE, newParameterName, data);
    }
  }

  private void defineParameter(
      final VerilogDefineParameter defineNode,
      final VerilogContext context) {
    InvariantChecks.checkNotNull(defineNode);
    InvariantChecks.checkNotNull(context);

    // Define path is considered to be relative, not absolute.
    final String newParameterName = context.getFullName(defineNode.getPath().toString());

    final Node oldExpression = defineNode.getExpression().getMinExpression().getNode();
    final Node newExpression = substitute(oldExpression, context.getBindings());

    final Map<String, VerilogParameter> parameters = designBuilder.getParameters();
    final Data data = evaluate(newExpression, context, parameters);
    InvariantChecks.checkNotNull(data, String.format("expression=%s", newExpression));

    VerilogLogger.debug("Define parameter: %s = %s = %s", newParameterName, newExpression, data);
    designBuilder.addParameter(VerilogParameter.Type.DEFPARAM, newParameterName, data);
  }

  private static Node substitute(
      final Node expression,
      final Map<String, VerilogBinding> bindings) {

    return Transformer.substitute(expression,
        new NodeProvider() {
          @Override
          public Node getNode(final Variable variable) {
            final VerilogBinding binding = bindings.get(variable.getName());
            return binding != null ? binding.getRhs() : null;
          }
        });
  }

  private static Data evaluate(
      final Node expression,
      final VerilogContext context,
      final Map<String, VerilogParameter> parameters) {

    return VerilogCalculator.evaluate(expression,
        new ValueProvider() {
          @Override
          public Data getVariableValue(final Variable variable) {
            final String name = variable.getName();

            // Search among generate variables.
            final Data value = context.getGenvarValue(name);
            if (value != null) {
              return value;
            }

            // Search among instance parameters.
            final VerilogParameter parameter = parameters.get(name);
            if (parameter != null) {
              return parameter.getValue();
            }

            // Only generate variable and parameters are in use.
            return null;
          }
        });
  }

  private static boolean evaluateBoolean(
      final Node expression,
      final VerilogContext context,
      final Map<String, VerilogParameter> parameters) {
    final Data data = evaluate(expression, context, parameters);
    InvariantChecks.checkNotNull(data, String.format("expression=%s", expression));

    if (data.isType(DataTypeId.LOGIC_BOOLEAN)) {
      return data.getBoolean();
    } else if (data.isType(DataTypeId.LOGIC_INTEGER)) {
      return data.getInteger().intValue() != 0;
    } else if (data.isType(DataTypeId.BIT_VECTOR)) {
      return data.getBitVector().intValue() != 0;
    } else {
      InvariantChecks.checkTrue(false, "Incorrect data type");
      return false;
    }
  }
}

