/*
 * Copyright 2018-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.elaborator;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.VerilogActivity;
import ru.ispras.verilog.parser.model.VerilogAssign;
import ru.ispras.verilog.parser.model.VerilogAssignStatement;
import ru.ispras.verilog.parser.model.VerilogAssignment;
import ru.ispras.verilog.parser.model.VerilogDeclaration;
import ru.ispras.verilog.parser.model.VerilogNode;
import ru.ispras.verilog.parser.model.VerilogStatement;
import ru.ispras.verilog.parser.model.basis.VerilogPath;
import ru.ispras.verilog.parser.model.basis.VerilogPathItem;
import ru.ispras.verilog.parser.model.basis.VerilogReference;
import ru.ispras.verilog.parser.transformer.VerilogTransformerBvconcat;
import ru.ispras.verilog.parser.transformer.VerilogTransformerComposite;
import ru.ispras.verilog.parser.transformer.VerilogTransformerOperation;
import ru.ispras.verilog.parser.transformer.VerilogTransformerVariableSubstitute;

import java.util.Map;

/**
 * {@link VerilogInstantiator} produces a process instance.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogInstantiator {
  private final VerilogProcess.Type type;
  private final VerilogNode node;
  private final VerilogContext context;

  public VerilogInstantiator(
      final VerilogActivity activity,
      final VerilogContext context) {
    InvariantChecks.checkNotNull(activity);
    InvariantChecks.checkNotNull(context);

    // Process statement.
    this.type = activity.isAlways() ? VerilogProcess.Type.ALWAYS : VerilogProcess.Type.INITIAL;
    this.node = activity.getStatement();
    this.context = context;
  }

  public VerilogInstantiator(
      final VerilogAssign assign,
      final VerilogContext context) {
    InvariantChecks.checkNotNull(assign);
    InvariantChecks.checkNotNull(context);

    final VerilogAssignStatement statement = new VerilogAssignStatement(assign.getParentNode());

    // Continuous assignment statement.
    statement.setAssign();
    statement.setAssignment(assign.getAssignment());

    this.type = VerilogProcess.Type.ASSIGN;
    this.node = statement;
    this.context = context;
  }

  public VerilogInstantiator(
      final VerilogDeclaration declaration,
      final VerilogContext context,
      final boolean toAssign) {
    InvariantChecks.checkNotNull(declaration);
    InvariantChecks.checkNotNull(context);

    if (toAssign) {
      // Continuous assignment statement.
      final VerilogNode parent = declaration.getParentNode();
      final VerilogAssignStatement statement = new VerilogAssignStatement(parent);

      // Continuous assignment statement.
      statement.setAssign();
      final VerilogAssignment assignment = new VerilogAssignment(parent);

      final VerilogReference reference = new VerilogReference();
      reference.setDeclaration(declaration);
      reference.setPath(new VerilogPath(new VerilogPathItem(declaration.getName())));

      assignment.addReference(reference);
      assignment.setRhsExpression(declaration.getExpression());

      statement.setAssignment(assignment);

      this.type = VerilogProcess.Type.ASSIGN;
      this.node = statement;

    } else {
      // Declaration.
      this.type = null;
      this.node = declaration;
    }

    this.context = context;
  }

  public VerilogProcess instantiateProcess(
      final Map<String, VerilogDescriptor> variables,
      final Map<String, VerilogParameter> parameters) {
    final VerilogStatement instance =
        (VerilogStatement) instantiate(node, variables, parameters);
    return new VerilogProcess(type, instance);
  }

  public VerilogDescriptor instantiateDescriptor(
      final VerilogVariable variable,
      final Map<String, VerilogDescriptor> variables,
      final Map<String, VerilogParameter> parameters) {
    final VerilogDeclaration instance =
        (VerilogDeclaration) instantiate(node, variables, parameters);
    return new VerilogDescriptor(variable, instance);
  }

  private VerilogNode instantiate(
      final VerilogNode node,
      final Map<String, VerilogDescriptor> variables,
      final Map<String, VerilogParameter> parameters) {
    InvariantChecks.checkNotNull(variables);
    InvariantChecks.checkNotNull(parameters);

    // Node is copied to avoid issues on consequent instantiations.
    final VerilogNode instance = node.clone();
    final VerilogTransformerComposite transformer = new VerilogTransformerComposite();

    // Substitute local variables with the global ones (taking into account bit ranges).
    transformer.add(new VerilogTransformerVariableSubstitute(variables, context, parameters));
    // Remove empty BVREPEATs from BVCONCATs.
    transformer.add(new VerilogTransformerBvconcat());
    // Perform general transformations.
    transformer.add(new VerilogTransformerOperation());
    transformer.run(instance);

    return instance;
  }
}
