/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.basis.VerilogEventControl;

import java.util.EnumSet;

/**
 * {@link VerilogAssignment} represents assignment statements.
 * 
 * <p>There are the following assignment statements:</p>
 * <ul>
 * <li>{@code =} (blocking);</li>
 * <li>{@code <=} (non-blocking);</li>
 * <li>{@code assign};</li>
 * <li>{@code deassign};</li>
 * <li>{@code force};</li>
 * <li>{@code release}.</li>
 * </ul>
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogAssignStatement extends VerilogStatement {
  public static final Tag TAG = Tag.ASSIGN_STATEMENT;
  public static final EnumSet<Tag> TAGS_CHILDREN = EnumSet.noneOf(Tag.class);

  /** {@link Type} contains the assignment statement types. */
  public enum Type {
    /** Blocking assignment: {@code r = e}. */
    BLOCKING,
    /** Non-blocking assignment: {@code r <= e}. */
    NON_BLOCKING,
    /** Continuous assignment: {@code assign r = e}. */
    ASSIGN,
    /** Deassignment: {@code deassign r}. */
    DEASSIGN,
    /** Force assignment: {@code force r = e}. */
    FORCE,
    /** Release: {@code release r}. */
    RELEASE
  }

  /** Type of the assignment statement (by default, {@code BLOCKING}). */
  private Type type = Type.BLOCKING;
  /** Assignment itself (a pair of l- and r-values). */
  private VerilogAssignment assignment;
  /** Assignment control (delay or event). */
  private VerilogEventControl control;

  /**
   * Creates an assignment statement.
   * 
   * @param parent the node parent.
   */
  public VerilogAssignStatement(final VerilogNode parent) {
    super(TAG, parent);
  }

  /**
   * Creates a copy of the assignment statement.
   * 
   * @param other the assignment statement to be copied.
   * @param parent the parent node.
   */
  public VerilogAssignStatement(final VerilogAssignStatement other, final VerilogNode parent) {
    super(other, parent);

    this.type = other.type;
    this.assignment = other.assignment != null ? other.assignment.clone() : null;
    this.control = other.control != null ? other.control.clone() : null;
  }

  /**
   * Returns the assignment statement type.
   * 
   * @return the assignment statement type.
   */
  public Type getType() {
    return type;
  }

  /**
   * Sets the assignment statement type.
   * 
   * @param type the type to be set.
   */
  public void setType(final Type type) {
    InvariantChecks.checkNotNull(type);
    this.type = type;
  }

  /**
   * Checks whether the statement is {@code BLOCKING}.
   * 
   * @return {@code true} iff the statement is {@code BLOCKING}.
   */
  public boolean isBlocking() {
    return type == Type.BLOCKING;
  }

  /**
   * Checks whether the statement is {@code NON_BLOCKING}.
   * 
   * @return {@code true} iff the statement is {@code NON_BLOCKING}.
   */
  public boolean isNonBlocking() {
    return type == Type.NON_BLOCKING;
  }

  /**
   * Checks whether the statement is {@code ASSIGN}.
   * 
   * @return {@code true} iff the statement is {@code ASSIGN}.
   */
  public boolean isAssign() {
    return type == Type.ASSIGN;
  }

  /**
   * Checks whether the statement is {@code DEASSIGN}.
   * 
   * @return {@code true} iff the statement is {@code DEASSIGN}.
   */
  public boolean isDeassign() {
    return type == Type.DEASSIGN;
  }

  /**
   * Checks whether the statement is {@code FORCE}.
   * 
   * @return {@code true} iff the statement is {@code FORCE}.
   */
  public boolean isForce() {
    return type == Type.FORCE;
  }

  /**
   * Checks whether the statement is {@code RELEASE}.
   * 
   * @return {@code true} iff the statement is {@code RELEASE}.
   */
  public boolean isRelease() {
    return type == Type.RELEASE;
  }

  /**
   *  Sets the type of the statement to {@code BLOCKING}.
   */
  public void setBlocking() {
    type = Type.BLOCKING;
  }

  /**
   * Sets the type of the statement to {@code NON_BLOCKING}.
   */
  public void setNonBlocking() {
    type = Type.NON_BLOCKING;
  }

  /**
   * Sets the type of the statement to {@code ASSIGN}.
   */
  public void setAssign() {
    type = Type.ASSIGN;
  }

  /**
   * Sets the type of the statement to {@code DEASSIGN}.
   */
  public void setDeassign() {
    type = Type.DEASSIGN;
  }

  /**
   * Sets the type of the statement to {@code FORCE}.
   */
  public void setForce() {
    type = Type.FORCE;
  }

  /**
   * Sets the type of the statement to {@code RELEASE}.
   */
  public void setRelease() {
    type = Type.RELEASE;
  }

  /**
   * Returns the assignment of the statement (the pair of l- and r-values).
   * 
   * @return the assignment.
   */
  public VerilogAssignment getAssignment() {
    return assignment;
  }

  /**
   * Sets the assignment of the statement (the pair of l- and r-values).
   * 
   * @param assignment the assignment to be set.
   */
  public void setAssignment(final VerilogAssignment assignment) {
    InvariantChecks.checkNotNull(assignment);
    this.assignment = assignment;
  }

  /**
   * Returns the control of the statement (the delay or event).
   * 
   * @return the control.
   */
  public VerilogEventControl getEventControl() {
    return control;
  }

  /**
   * Sets the control of the statement (the delay or event).
   * 
   * @param control the control to be set.
   */
  public void setEventControl(VerilogEventControl control) {
    InvariantChecks.checkNotNull(control);
    this.control = control;
  }

  @Override
  public VerilogAssignStatement clone() {
    return new VerilogAssignStatement(this, getParentNode());
  }
}
