/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model;

import ru.ispras.fortress.util.InvariantChecks;

import java.util.EnumSet;

/**
 * {@link VerilogBlockStatement} represents block statements.
 * 
 * <p>There are the following block types:</p>
 * <ul>
 * <li>{@code begin} (sequential block);</li>
 * <li>{@code fork} (parallel block).</li>
 * </ul>
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogBlockStatement extends VerilogStatement {
  public static final Tag TAG = Tag.BLOCK_STATEMENT;
  public static final EnumSet<Tag> TAGS_ELEMENTS = EnumSet.of(Tag.DECLARATION);
  public static final EnumSet<Tag> TAGS_CHILDREN = union(VerilogStatement.TAGS, TAGS_ELEMENTS);

  /** {@link Type} contains the block statement types. */
  public enum Type {
    /** Sequential block: {@code begin ... end}. */
    BEGIN,
    /** Parallel block: {@code fork ... end}. */
    FORK
  }

  /** Type of the block statement (by default, {@code BEGIN}). */
  private Type type = Type.BEGIN;

  /**
   * Creates a block statement.
   * 
   * @param parent the parent node.
   */
  public VerilogBlockStatement(final VerilogNode parent) {
    super(TAG, TAGS_CHILDREN, NodeKind.NAMED_BLOCK, parent);
  }

  /**
   * Creates a copy of the block statement.
   * 
   * @param other the block statement to be copied.
   * @param parent the parent node.
   */
  public VerilogBlockStatement(final VerilogBlockStatement other, final VerilogNode parent) {
    super(other, parent);

    this.type = other.type;
  }

  /**
   * Returns the type of the block statement ({@code BEGIN} or {@code FORK}).
   * 
   * @return the type of the block statement.
   */
  public Type getType() {
    return type;
  }

  /**
   * Sets the type of the block statement.
   * 
   * @param type the type to be set.
   */
  public void setType(final Type type) {
    InvariantChecks.checkNotNull(type);
    this.type = type;
  }

  /**
   * Checks whether the statement {@code BEGIN}.
   * 
   * @return {@code true} iff the statement is {@code BEGIN}.
   */
  public boolean isBegin() {
    return type == Type.BEGIN;
  }

  /**
   * Checks whether the statement {@code FORK}.
   * 
   * @return {@code true} iff the statement is {@code FORK}.
   */
  public boolean isFork() {
    return type == Type.FORK;
  }

  /**
   * Sets the type of the statement to {@code BEGIN}.
   */
  public void setBegin() {
    type = Type.BEGIN;
  }

  /**
   * Sets the type of the statement to {@code FORK}.
   */
  public void setFork() {
    type = Type.FORK;
  }

  @Override
  public VerilogBlockStatement clone() {
    return new VerilogBlockStatement(this, getParentNode());
  }
}
