/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.core.AbstractNode;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.EnumSet;
import java.util.List;

/**
 * {@link VerilogCaseGenerate} represents generate case constructs.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogCaseGenerate extends VerilogNode {
  public static final Tag TAG = Tag.CASE_GENERATE;
  public static final EnumSet<Tag> TAGS_CHILDREN = EnumSet.of(Tag.CASE_GENERATE_ITEM);

  /** Case selection expression (to be matched with available cases). */
  private VerilogExpression expression;

  /**
   * Creates a generate case construct.
   * 
   * @param parent the parent node.
   */
  public VerilogCaseGenerate(final VerilogNode parent) {
    super(TAG, TAGS_CHILDREN, NodeKind.PSEUDO_ELEMENT, parent);
  }

  /**
   * Creates a copy of the generate case construct.
   * 
   * @param other the generate case construct to be copied.
   * @param parent the parent node.
   */
  public VerilogCaseGenerate(final VerilogCaseGenerate other, final VerilogNode parent) {
    super(other, parent);

    this.expression = other.expression != null ? other.expression.clone() : null;
  }
  
  /**
   * Returns the case selection expression.
   * 
   * @return the expression.
   */
  public VerilogExpression getExpression() {
    return expression;
  }

  /**
   * Sets the case selection expression.
   * 
   * @param expression the expression to be set.
   */
  public void setExpression(final VerilogExpression expression) {
    InvariantChecks.checkNotNull(expression);
    this.expression = expression;
  }

  /**
   * Returns the cases.
   * 
   * @return the list of cases.
   */
  public List<VerilogCaseGenerateItem> getCases() {
    final Collection<AbstractNode<VerilogNode.Tag>> cases = items(TAGS_CHILDREN);
    return new ArrayList<>(Arrays.asList(cases.toArray(new VerilogCaseGenerateItem[]{})));
  }

  /**
   * Adds the case to the case selection construct.
   * 
   * @param item the case to be added.
   */
  public void addCase(final VerilogCaseGenerateItem item) {
    add(item);
  }

  @Override
  public VerilogCaseGenerate clone() {
    return new VerilogCaseGenerate(this, getParentNode());
  }
}
