/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.basis.VerilogDelay;
import ru.ispras.verilog.parser.model.basis.VerilogRange;
import ru.ispras.verilog.parser.model.basis.VerilogStrength;

import java.util.EnumSet;

/**
 * {@link VerilogInstantiation} represents module instantiations.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogInstantiation extends VerilogNode {
  public static final Tag TAG = Tag.INSTANTIATION;
  public static final EnumSet<Tag> TAGS_CHILDREN = EnumSet.of(
      // Instance parameters.
      Tag.ASSIGNMENT,
      // Port connections.
      Tag.PORT_CONNECTION
  );

  /** Module name. */
  private String module;
  /** Port connection delay. */
  private VerilogDelay delay;
  /** Port connection strength. */
  private VerilogStrength strength;
  /** Instantiation array range. */
  private VerilogRange range;
  /** Module declaration. */
  private VerilogModule declaration;

  /**
   * Creates an instantiation.
   * 
   * @param parent the parent node.
   */
  public VerilogInstantiation(final VerilogNode parent) {
    super(TAG, TAGS_CHILDREN, NodeKind.OBJECT, parent);
  }

  /**
   * Creates a copy of the instantiation.
   * 
   * @param other the instantiation to be copied.
   * @param parent the parent node.
   */
  public VerilogInstantiation(final VerilogInstantiation other, final VerilogNode parent) {
    super(other, parent);

    this.module = other.module;
    this.delay = other.delay != null ? other.delay.clone() : null;
    this.strength = other.strength != null ? other.strength.clone() : null;
    this.range = other.range != null ? other.range.clone() : null;

    // There is no need to clone the declaration.
    this.declaration = other.declaration;
  }

  /**
   * Returns the module name of the instance.
   * 
   * @return the module name.
   */
  public String getModuleName() {
    return module;
  }

  public void setModuleName(final String module) {
    InvariantChecks.checkNotNull(module);
    this.module = module;
  }

  /**
   * Returns the delay of the instance's port connections.
   * 
   * @return the delay.
   */
  public VerilogDelay getDelay() {
    return delay;
  }

  /**
   * Sets the delay of the instance's port connections.
   * 
   * @param delay the delay to be set.
   */
  public void setDelay(final VerilogDelay delay) {
    InvariantChecks.checkNotNull(delay);
    this.delay = delay;
  }

  /**
   * Returns the strength of the instance's port connections.
   * 
   * @return the strength.
   */
  public VerilogStrength getStrength() {
    return strength;
  }

  /**
   * Sets the strength of the instance's port connections.
   * 
   * @param strength the strength to be set.
   */
  public void setStrength(final VerilogStrength strength) {
    InvariantChecks.checkNotNull(strength);
    this.strength = strength;
  }

  /**
   * Returns the range of the instantiation array.
   * 
   * @return the range.
   */
  public VerilogRange getRange() {
    return range;
  }

  /**
   * Sets the range of the instantiation array.
   * 
   * @param range the range to be set.
   */
  public void setRange(final VerilogRange range) {
    InvariantChecks.checkNotNull(range);
    this.range = range;
  }

  /**
   * Adds the parameter definition to the instance.
   * 
   * @param assignment the parameter definition.
   */
  public void addParameter(final VerilogAssignment assignment) {
    add(assignment);
  }

  /**
   * Adds a port connection to the instance.
   * 
   * @param connection the port connection.
   */
  public void addConnection(final VerilogPortConnection connection) {
    add(connection);
  }

  /**
   * Returns the module declaration.
   * 
   * @return the module declaration.
   */
  public VerilogModule getDeclaration() {
    return declaration;
  }

  /**
   * Sets the module declaration.
   * 
   * @param declaration the module declaration to be set.
   */
  public void setDeclaration(final VerilogModule declaration) {
    InvariantChecks.checkNotNull(declaration);
    this.declaration = declaration;
  }

  @Override
  public VerilogInstantiation clone() {
    return new VerilogInstantiation(this, getParentNode());
  }
}
