/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model;

import ru.ispras.fortress.util.InvariantChecks;

import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;

/**
 * {@link VerilogTableEntry} represents UDP table entries.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogTableEntry extends VerilogNode {
  public static final Tag TAG = Tag.TABLE_ENTRY;
  public static final EnumSet<Tag> TAGS_CHILDREN = EnumSet.noneOf(Tag.class);

  /** Input signals 1. */
  private List<String> inputs1 = new ArrayList<>();
  /** Input signals 2. */
  private List<String> inputs2 = new ArrayList<>();
  /** Edges. */
  private List<String> edge = new ArrayList<>();
  /** Output signals. */
  private List<String> outputs = new ArrayList<>();

  /**
   * Creates a UDP table entry.
   * 
   * @param parent the parent node.
   */
  public VerilogTableEntry(final VerilogNode parent) {
    super(TAG, TAGS_CHILDREN, NodeKind.OBJECT, parent);
  }

  /**
   * Creates a copy of the UDP table entry.
   * 
   * @param other the UDP table entry,
   * @param parent the parent node.
   */
  public VerilogTableEntry(final VerilogTableEntry other, final VerilogNode parent) {
    super(other, parent);

    this.inputs1 = new ArrayList<>(other.inputs1);
    this.inputs2 = new ArrayList<>(other.inputs2);
    this.edge = new ArrayList<>(other.edge);
    this.outputs = new ArrayList<>(other.outputs);
  }

  /**
   * Returns the input signals (1).
   * 
   * @return the list of input signals.
   */
  public List<String> getInputs1() {
    return inputs1;
  }

  /**
   * Adds the input to the list (1).
   * 
   * @param input1 the input to be added.
   */
  public void addInput1(final String input1) {
    InvariantChecks.checkNotNull(input1);
    inputs1.add(input1);
  }

  /**
   * Returns the input signals (2).
   * 
   * @return the list of input signals.
   */
  public List<String> getInputs2() {
    return inputs2;
  }

  /**
   * Adds the input to the list (2).
   * 
   * @param input2 the input to be added.
   */
  public void addInput2(final String input2) {
    InvariantChecks.checkNotNull(input2);
    inputs2.add(input2);
  }

  /**
   * Returns the edges.
   * 
   * @return the list of edges.
   */
  public List<String> getEdge() {
    return edge;
  }

  /**
   * Adds the edge to the list.
   * 
   * @param edge the edge to be added.
   */
  public void addEdge(final String edge) {
    InvariantChecks.checkNotNull(edge);
    this.edge.add(edge);
  }

  /**
   * Returns the output signals.
   * 
   * @return the list of output signals.
   */
  public List<String> getOutputs() {
    return outputs;
  }

  /**
   * Adds the output to the list.
   * 
   * @param output the output to be added.
   */
  public void addOutput(final String output) {
    InvariantChecks.checkNotNull(output);
    outputs.add(output);
  }

  @Override
  public VerilogTableEntry clone() {
    return new VerilogTableEntry(this, getParentNode());
  }
}
