/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;
import ru.ispras.verilog.parser.model.basis.VerilogPath;

import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;

/**
 * {@link VerilogTaskStatement} represents task statements.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogTaskStatement extends VerilogStatement {
  public static final Tag TAG = Tag.TASK_STATEMENT;
  public static final EnumSet<Tag> TAGS_CHILDREN = EnumSet.noneOf(Tag.class);

  /** Path to the task. */
  private VerilogPath path;
  /** Arguments of the task. */
  private List<VerilogExpression> args = new ArrayList<>();
  /** Declaration of the task. */
  private VerilogProcedure declaration;

  /**
   * Creates a task statement.
   * 
   * @param parent the parent node.
   */
  public VerilogTaskStatement(final VerilogNode parent) {
    super(TAG, parent);
  }

  public VerilogTaskStatement(final VerilogTaskStatement other, final VerilogNode parent) {
    super(other, parent);

    this.path = other.path != null ? other.path.clone() : null;

    for (final VerilogExpression expression : other.args) {
      this.args.add(expression.clone());
    }

    this.declaration = other.declaration != null ? other.declaration.clone() : null;
  }

  /**
   * Returns the path to the task.
   * 
   * @return the task path.
   */
  public VerilogPath getPath() {
    return path;
  }

  /**
   * Sets the path of the task.
   * 
   * @param path the path to be set.
   */
  public void setPath(final VerilogPath path) {
    InvariantChecks.checkNotNull(path);
    this.path = path;
  }

  /**
   * Returns the arguments of the task statement.
   * 
   * @return the arguments.
   */
  public List<VerilogExpression> getArguments() {
    return args;
  }

  /**
   * Adds the argument to the task statement.
   * 
   * @param argument the argument to be added.
   */
  public void addArgument(final VerilogExpression argument) {
    InvariantChecks.checkNotNull(argument);
    args.add(argument);
  }

  /**
   * Returns the task declaration.
   * 
   * @return the task declaration.
   */
  public VerilogProcedure getDeclaration() {
    return declaration;
  }

  /**
   * Sets the task declaration.
   * 
   * @param declaration the task declaration to be set.
   */
  public void setDeclaration(final VerilogProcedure declaration) {
    InvariantChecks.checkNotNull(declaration);
    this.declaration = declaration;
  }

  @Override
  public VerilogTaskStatement clone() {
    return new VerilogTaskStatement(this, getParentNode());
  }
}
