/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.model.basis;

import ru.ispras.fortress.util.InvariantChecks;

/**
 * {@link VerilogRange} represents ranges.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogRange {

  /** {@link Type} contains the range types. */
  public enum Type {
    /** Usual range: {@code [upper : lower]}. */
    RANGE,
    /** Plus range: {@code [lower +: width]}. */
    PLUS,
    /** Minus range: {@code [upper -: width]}. */
    MINUS
  }

  /** Type of the range. */
  private Type type = Type.RANGE;
  /** Left expression. */
  private VerilogExpression lhs;
  /** Right expression. */
  private VerilogExpression rhs;

  /**
   * Creates an range.
   */
  public VerilogRange() {}

  /**
   * Creates a copy of the range.
   * 
   * @param other the range to be copied.
   */
  public VerilogRange(final VerilogRange other) {
    InvariantChecks.checkNotNull(other);

    this.type = other.type;
    this.lhs = other.lhs != null ? other.lhs.clone() : null;
    this.rhs = other.rhs != null ? other.rhs.clone() : null;
  }

  /**
   * Returns the type of the range.
   *
   * @return the range type.
   */
  public Type getType() {
    return type;
  }

  /**
   * Sets the type of the range.
   *
   * @param type the type to be set.
   */
  public void setType(final Type type) {
    InvariantChecks.checkNotNull(type);
    this.type = type;
  }

  /**
   * Checks whether the range is {@code RANGE}.
   *
   * @return {@code true} iff the range is {@code RANGE}.
   */
  public boolean isRange() {
    return type == Type.RANGE;
  }

  /**
   * Checks whether the range is {@code PLUS}.
   *
   * @return {@code true} iff the range is {@code PLUS}.
   */
  public boolean isPlus() {
    return type == Type.PLUS;
  }

  /**
   * Checks whether the range is {@code MINUS}.
   *
   * @return {@code true} if the range is {@code MINUS}.
   */
  public boolean isMinus() {
    return type == Type.MINUS;
  }

  /**
   * Sets the range type to {@code RANGE}.
   */
  public void setRange() {
    type = Type.RANGE;
  }

  /**
   * Sets the range type to {@code PLUS}.
   */
  public void setPlus() {
    type = Type.PLUS;
  }

  /**
   * Sets the range type to {@code MINUS}.
   */
  public void setMinus() {
    type = Type.MINUS;
  }

  /**
   * Returns the left expression of the range.
   *
   * @return the left expression.
   */
  public VerilogExpression getLeftExpression() {
    return lhs;
  }

  /**
   * Sets the left expression of the range.
   *
   * @param lhs the left expression to be set.
   */
  public void setLeftExpression(final VerilogExpression lhs) {
    InvariantChecks.checkNotNull(lhs);
    this.lhs = lhs;
  }

  /**
   * Returns the right expression of the range.
   *
   * @return the right expression.
   */
  public VerilogExpression getRightExpression() {
    return rhs;
  }

  /**
   * Sets the right expression of the range.
   *
   * @param rhs the right expression to be set.
   */
  public void setRightExpression(final VerilogExpression rhs) {
    InvariantChecks.checkNotNull(rhs);
    this.rhs = rhs;
  }

  @Override
  public String toString() {
    if (rhs == null) {
      return String.format("[%s]", lhs);
    }

    switch(type) {
      case PLUS:
        return String.format("[%s+:%s]", lhs, rhs);
      case MINUS:
        return String.format("[%s+:%s]", lhs, rhs);
      default:
        return String.format("X[%s:%s]",  lhs, rhs);
    }
  }

  @Override
  public VerilogRange clone() {
    return new VerilogRange(this);
  }
}
