/*
 * Copyright 2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.transformer;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.model.VerilogAssign;
import ru.ispras.verilog.parser.model.VerilogAssignStatement;
import ru.ispras.verilog.parser.model.VerilogAssignment;
import ru.ispras.verilog.parser.model.VerilogCaseStatement;
import ru.ispras.verilog.parser.model.VerilogCaseStatementItem;
import ru.ispras.verilog.parser.model.VerilogDeclaration;
import ru.ispras.verilog.parser.model.VerilogDelayedStatement;
import ru.ispras.verilog.parser.model.VerilogIfStatement;
import ru.ispras.verilog.parser.model.VerilogLoopStatement;
import ru.ispras.verilog.parser.model.VerilogNode;
import ru.ispras.verilog.parser.model.VerilogTaskStatement;
import ru.ispras.verilog.parser.model.VerilogWaitStatement;
import ru.ispras.verilog.parser.model.basis.VerilogDelay;
import ru.ispras.verilog.parser.model.basis.VerilogElementType;
import ru.ispras.verilog.parser.model.basis.VerilogEvent;
import ru.ispras.verilog.parser.model.basis.VerilogEventControl;
import ru.ispras.verilog.parser.model.basis.VerilogExpression;
import ru.ispras.verilog.parser.model.basis.VerilogMinTypMax;
import ru.ispras.verilog.parser.model.basis.VerilogRange;
import ru.ispras.verilog.parser.model.basis.VerilogReference;
import ru.ispras.verilog.parser.walker.VerilogEmptyVisitor;
import ru.ispras.verilog.parser.walker.VerilogTreeWalker;

import java.util.List;

/**
 * {@link VerilogTransformer} implements an engine that transforms expressions.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public abstract class VerilogTransformer extends VerilogEmptyVisitor {
  public final void run(final VerilogNode node) {
    final VerilogTreeWalker walker = new VerilogTreeWalker(node, this);
    walker.start();
  }

  @Override
  public Result onDeclarationBegin(final VerilogDeclaration node) {
    final VerilogElementType elementType = node.getElementType();
    final VerilogMinTypMax expression = node.getExpression();

    for (final VerilogRange dimension : node.getDimensions()) {
      transform(dimension);
    }

    transform(elementType);

    if (expression != null) {
      transform(expression);
    }

    return Result.OK;
  }

  @Override
  public Result onAssignBegin(final VerilogAssign node) {
    transform(node.getAssignment());
    return Result.OK;
  }

  @Override
  public Result onAssignStatementBegin(final VerilogAssignStatement node) {
    transform(node.getAssignment());
    return Result.OK;
  }

  @Override
  public Result onCaseStatementBegin(final VerilogCaseStatement node) {
    transform(node.getExpression());
    return Result.OK;
  }

  @Override
  public Result onCaseStatementItemBegin(final VerilogCaseStatementItem node) {
    for (final VerilogExpression expression: node.getExpressions()) {
      transform(expression);
    }
    return Result.OK;
  }

  @Override
  public Result onDelayedStatementBegin(final VerilogDelayedStatement node) {
    transform(node.getEventControl());
    return Result.OK;
  }

  @Override
  public Result onIfStatementBegin(final VerilogIfStatement node) {
    transform(node.getExpression());
    return Result.OK;
  }

  @Override
  public Result onLoopStatementBegin(final VerilogLoopStatement node) {
    final VerilogAssignment initialization = node.getInitialization();
    final VerilogExpression expression = node.getExpression();
    final VerilogAssignment iteration = node.getIteration();

    if (initialization != null) {
      transform(initialization);
    }
    if (expression != null) {
      transform(expression);
    }
    if (iteration != null) {
      transform(iteration);
    }

    return Result.OK;
  }

  @Override
  public Result onTaskStatementBegin(final VerilogTaskStatement node) {
    for (final VerilogExpression expression : node.getArguments()) {
      transform(expression);
    }

    return Result.OK;
  }

  @Override
  public Result onWaitStatementBegin(final VerilogWaitStatement node) {
    transform(node.getExpression());
    return Result.OK;
  }

  private void transform(final VerilogElementType node) {
    InvariantChecks.checkNotNull(node);

    final VerilogRange range = node.getRange();

    if (range != null) {
      transform(range);
    }
  }

  private void transform(final VerilogRange range) {
    final VerilogExpression lhs = range.getLeftExpression();
    final VerilogExpression rhs = range.getRightExpression();

    if (lhs != null) {
      transform(lhs);
    }
    if (rhs != null) {
      transform(rhs);
    }
  }

  private void transform(final VerilogReference node) {
    if (node.getNode() != null) {
      final Node newNode = transform(node.getNode());
      node.setNode(newNode);
    }
  }

  private void transform(final VerilogExpression node) {
    InvariantChecks.checkNotNull(node);

    final Node newNode = transform(node.getNode());
    node.setNode(newNode);
  }

  private void transform(final VerilogMinTypMax node) {
    InvariantChecks.checkNotNull(node);

    final VerilogExpression min = node.getMinExpression();
    final VerilogExpression typ = node.getTypExpression();
    final VerilogExpression max = node.getMaxExpression();

    if (min != null) {
      transform(min);
    }
    if (typ != null) {
      transform(typ);
    }
    if (max != null) {
      transform(max);
    }
  }

  private void transform(final VerilogAssignment node) {
    InvariantChecks.checkNotNull(node);

    for (final VerilogReference reference : node.getReferences()) {
      transform(reference);
    }

    final VerilogMinTypMax expression = node.getRhsExpression();
    if (expression != null) {
      transform(expression);
    }
  }

  private void transform(final VerilogDelay node) {
    InvariantChecks.checkNotNull(node);

    for (final VerilogMinTypMax expression : node.getDelays()) {
      transform(expression);
    }
  }

  private void transform(final VerilogEventControl node) {
    InvariantChecks.checkNotNull(node);

    final VerilogDelay delay = node.getDelay();
    final List<VerilogEvent> events = node.getEvents();
    final VerilogExpression repeat = node.getRepeat();

    if (delay != null) {
      transform(delay);
    }
    if (events != null) {
      for (final VerilogEvent event: node.getEvents()) {
        transform(event.getExpression());
      }
    }
    if (repeat != null) {
      transform(repeat);
    }
  }

  public abstract Node transform(final Node node);
}
