/*
 * Copyright 2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.verilog.parser.transformer;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.transformer.TransformerRule;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.verilog.parser.interpreter.VerilogCalculator;
import ru.ispras.verilog.parser.model.VerilogNode;

import java.util.ArrayList;
import java.util.List;

/**
 * {@link VerilogTransformerBvconcat} removes empty {@code BVREPEAT} from {@code BVCONCAT}.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class VerilogTransformerBvconcat extends VerilogTransformer {

  /**
   * Reduces the expression and applies type casting
   * 
   * @param node the AST node to be transformed.
   */
  public static void transform(final VerilogNode node) {
    InvariantChecks.checkNotNull(node);

    final VerilogTransformer transformer = new VerilogTransformerBvconcat();
    transformer.run(node);
  }

  @Override
  public Node transform(final Node node) {
    InvariantChecks.checkNotNull(node);

    return Transformer.transform(node, StandardOperation.BVCONCAT, new TransformerRule() {
      @Override
      public boolean isApplicable(final Node node) {
        return ExprUtils.isOperation(node, StandardOperation.BVCONCAT);
      }

      @Override
      public Node apply(final Node node) {
        final NodeOperation oldOperation = (NodeOperation) node;
        final List<Node> newOperands = new ArrayList<>(oldOperation.getOperandCount());

        for (int i = 0; i < oldOperation.getOperandCount(); i++) {
          final Node oldOperand = oldOperation.getOperand(i);

          if (ExprUtils.isOperation(oldOperand, StandardOperation.BVREPEAT)) {
            final NodeOperation repeat = (NodeOperation) oldOperand;
            final Integer count = VerilogCalculator.evaluateInteger(repeat.getOperand(0));

            if (count == null || count > 0) {
              newOperands.add(oldOperand);
            }
          } else {
            newOperands.add(oldOperand);
          }
        }

        return Nodes.bvconcat(newOperands);
      }
    });
  }
}
