/*
 * Copyright 2013-2018 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.verilog.parser.walker;

import ru.ispras.verilog.parser.core.AbstractNode;
import ru.ispras.verilog.parser.core.NodeVisitor;
import ru.ispras.verilog.parser.model.VerilogActivity;
import ru.ispras.verilog.parser.model.VerilogAssign;
import ru.ispras.verilog.parser.model.VerilogAssignStatement;
import ru.ispras.verilog.parser.model.VerilogAssignment;
import ru.ispras.verilog.parser.model.VerilogAttribute;
import ru.ispras.verilog.parser.model.VerilogBlockGenerate;
import ru.ispras.verilog.parser.model.VerilogBlockStatement;
import ru.ispras.verilog.parser.model.VerilogCaseGenerate;
import ru.ispras.verilog.parser.model.VerilogCaseGenerateItem;
import ru.ispras.verilog.parser.model.VerilogCaseStatement;
import ru.ispras.verilog.parser.model.VerilogCaseStatementItem;
import ru.ispras.verilog.parser.model.VerilogCode;
import ru.ispras.verilog.parser.model.VerilogDeclaration;
import ru.ispras.verilog.parser.model.VerilogDefineParameter;
import ru.ispras.verilog.parser.model.VerilogDelayedStatement;
import ru.ispras.verilog.parser.model.VerilogDisableStatement;
import ru.ispras.verilog.parser.model.VerilogGenerate;
import ru.ispras.verilog.parser.model.VerilogIfGenerate;
import ru.ispras.verilog.parser.model.VerilogIfGenerateBranch;
import ru.ispras.verilog.parser.model.VerilogIfStatement;
import ru.ispras.verilog.parser.model.VerilogIfStatementBranch;
import ru.ispras.verilog.parser.model.VerilogInstantiation;
import ru.ispras.verilog.parser.model.VerilogLoopGenerate;
import ru.ispras.verilog.parser.model.VerilogLoopStatement;
import ru.ispras.verilog.parser.model.VerilogModule;
import ru.ispras.verilog.parser.model.VerilogNode;
import ru.ispras.verilog.parser.model.VerilogNullStatement;
import ru.ispras.verilog.parser.model.VerilogPathDeclaration;
import ru.ispras.verilog.parser.model.VerilogPort;
import ru.ispras.verilog.parser.model.VerilogPortConnection;
import ru.ispras.verilog.parser.model.VerilogProcedure;
import ru.ispras.verilog.parser.model.VerilogPulseStyle;
import ru.ispras.verilog.parser.model.VerilogShowCancelled;
import ru.ispras.verilog.parser.model.VerilogSpecify;
import ru.ispras.verilog.parser.model.VerilogTable;
import ru.ispras.verilog.parser.model.VerilogTableEntry;
import ru.ispras.verilog.parser.model.VerilogTaskStatement;
import ru.ispras.verilog.parser.model.VerilogTriggerStatement;
import ru.ispras.verilog.parser.model.VerilogWaitStatement;

import java.util.HashMap;

/**
 * This class represents a base Verilog node visitor.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public abstract class VerilogNodeVisitor implements NodeVisitor<VerilogNode.Tag> {
  //------------------------------------------------------------------------------------------------
  // Node visitors to be override by a user
  //------------------------------------------------------------------------------------------------

  public abstract Result onActivityBegin(final VerilogActivity node);

  public abstract Result onActivityEnd(final VerilogActivity node);

  public abstract Result onAssignBegin(final VerilogAssign node);

  public abstract Result onAssignEnd(final VerilogAssign node);

  public abstract Result onAssignStatementBegin(final VerilogAssignStatement node);

  public abstract Result onAssignStatementEnd(final VerilogAssignStatement node);

  public abstract Result onAssignmentBegin(final VerilogAssignment node);

  public abstract Result onAssignmentEnd(final VerilogAssignment node);

  public abstract Result onAttributeBegin(final VerilogAttribute node);

  public abstract Result onAttributeEnd(final VerilogAttribute node);

  public abstract Result onBlockGenerateBegin(final VerilogBlockGenerate node);

  public abstract Result onBlockGenerateEnd(final VerilogBlockGenerate node);

  public abstract Result onBlockStatementBegin(final VerilogBlockStatement node);

  public abstract Result onBlockStatementEnd(final VerilogBlockStatement node);

  public abstract Result onCaseGenerateBegin(final VerilogCaseGenerate node);

  public abstract Result onCaseGenerateEnd(final VerilogCaseGenerate node);

  public abstract Result onCaseGenerateItemBegin(final VerilogCaseGenerateItem node);

  public abstract Result onCaseGenerateItemEnd(final VerilogCaseGenerateItem node);

  public abstract Result onCaseStatementBegin(final VerilogCaseStatement node);

  public abstract Result onCaseStatementEnd(final VerilogCaseStatement node);

  public abstract Result onCaseStatementItemBegin(final VerilogCaseStatementItem node);

  public abstract Result onCaseStatementItemEnd(final VerilogCaseStatementItem node);

  public abstract Result onCodeBegin(final VerilogCode node);

  public abstract Result onCodeEnd(final VerilogCode node);

  public abstract Result onDeclarationBegin(final VerilogDeclaration node);

  public abstract Result onDeclarationEnd(final VerilogDeclaration node);

  public abstract Result onDefineParameterBegin(final VerilogDefineParameter node);

  public abstract Result onDefineParameterEnd(final VerilogDefineParameter node);

  public abstract Result onDelayedStatementBegin(final VerilogDelayedStatement node);

  public abstract Result onDelayedStatementEnd(final VerilogDelayedStatement node);

  public abstract Result onDisableStatementBegin(final VerilogDisableStatement node);

  public abstract Result onDisableStatementEnd(final VerilogDisableStatement node);

  public abstract Result onGenerateBegin(final VerilogGenerate node);

  public abstract Result onGenerateEnd(final VerilogGenerate node);

  public abstract Result onIfGenerateBegin(final VerilogIfGenerate node);

  public abstract Result onIfGenerateEnd(final VerilogIfGenerate node);

  public abstract Result onIfGenerateBranchBegin(final VerilogIfGenerateBranch node);

  public abstract Result onIfGenerateBranchEnd(final VerilogIfGenerateBranch node);

  public abstract Result onIfStatementBegin(final VerilogIfStatement node);

  public abstract Result onIfStatementEnd(final VerilogIfStatement node);

  public abstract Result onIfStatementBranchBegin(final VerilogIfStatementBranch node);

  public abstract Result onIfStatementBranchEnd(final VerilogIfStatementBranch node);

  public abstract Result onInstantiationBegin(final VerilogInstantiation node);

  public abstract Result onInstantiationEnd(final VerilogInstantiation node);

  public abstract Result onLoopGenerateBegin(final VerilogLoopGenerate node);

  public abstract Result onLoopGenerateEnd(final VerilogLoopGenerate node);

  public abstract Result onLoopStatementBegin(final VerilogLoopStatement node);

  public abstract Result onLoopStatementEnd(final VerilogLoopStatement node);

  public abstract Result onModuleBegin(final VerilogModule node);

  public abstract Result onModuleEnd(final VerilogModule node);

  public abstract Result onNullStatementBegin(final VerilogNullStatement node);

  public abstract Result onNullStatementEnd(final VerilogNullStatement node);

  public abstract Result onPathDeclarationBegin(final VerilogPathDeclaration node);

  public abstract Result onPathDeclarationEnd(final VerilogPathDeclaration node);

  public abstract Result onPortBegin(final VerilogPort node);

  public abstract Result onPortEnd(final VerilogPort node);

  public abstract Result onPortConnectionBegin(final VerilogPortConnection node);

  public abstract Result onPortConnectionEnd(final VerilogPortConnection node);

  public abstract Result onProcedureBegin(final VerilogProcedure node);

  public abstract Result onProcedureEnd(final VerilogProcedure node);

  public abstract Result onPulseStyleBegin(final VerilogPulseStyle node);

  public abstract Result onPulseStyleEnd(final VerilogPulseStyle node);

  public abstract Result onShowCancelledBegin(final VerilogShowCancelled node);

  public abstract Result onShowCancelledEnd(final VerilogShowCancelled node);

  public abstract Result onSpecifyBegin(final VerilogSpecify node);

  public abstract Result onSpecifyEnd(final VerilogSpecify node);

  public abstract Result onTableBegin(final VerilogTable node);

  public abstract Result onTableEnd(final VerilogTable node);

  public abstract Result onTableEntryBegin(final VerilogTableEntry node);

  public abstract Result onTableEntryEnd(final VerilogTableEntry node);

  public abstract Result onTaskStatementBegin(final VerilogTaskStatement node);

  public abstract Result onTaskStatementEnd(final VerilogTaskStatement node);

  public abstract Result onTriggerStatementBegin(final VerilogTriggerStatement node);

  public abstract Result onTriggerStatementEnd(final VerilogTriggerStatement node);

  public abstract Result onWaitStatementBegin(final VerilogWaitStatement node);

  public abstract Result onWaitStatementEnd(final VerilogWaitStatement node);

  //------------------------------------------------------------------------------------------------
  // Internal structures
  //------------------------------------------------------------------------------------------------

  // An internal interface of a Verilog node handler.
  private interface VerilogNodeHandler {
    public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node);

    public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node);
  }

  // Handlers of Verilog nodes.
  private static HashMap<VerilogNode.Tag, VerilogNodeHandler> handlers = new HashMap<>();

  static {
    handlers.put(VerilogNode.Tag.ACTIVITY, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onActivityBegin((VerilogActivity) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onActivityEnd((VerilogActivity) node);
      }
    });

    handlers.put(VerilogNode.Tag.ASSIGN, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onAssignBegin((VerilogAssign) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onAssignEnd((VerilogAssign) node);
      }
    });

    handlers.put(VerilogNode.Tag.ASSIGN_STATEMENT, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onAssignStatementBegin((VerilogAssignStatement) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onAssignStatementEnd((VerilogAssignStatement) node);
      }
    });

    handlers.put(VerilogNode.Tag.ASSIGNMENT, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onAssignmentBegin((VerilogAssignment) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onAssignmentEnd((VerilogAssignment) node);
      }
    });

    handlers.put(VerilogNode.Tag.ATTRIBUTE, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onAttributeBegin((VerilogAttribute) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onAttributeEnd((VerilogAttribute) node);
      }
    });

    handlers.put(VerilogNode.Tag.BLOCK_GENERATE, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onBlockGenerateBegin((VerilogBlockGenerate) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onBlockGenerateEnd((VerilogBlockGenerate) node);
      }
    });

    handlers.put(VerilogNode.Tag.BLOCK_STATEMENT, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onBlockStatementBegin((VerilogBlockStatement) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onBlockStatementEnd((VerilogBlockStatement) node);
      }
    });

    handlers.put(VerilogNode.Tag.CASE_GENERATE, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onCaseGenerateBegin((VerilogCaseGenerate) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onCaseGenerateBegin((VerilogCaseGenerate) node);
      }
    });

    handlers.put(VerilogNode.Tag.CASE_GENERATE_ITEM, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onCaseGenerateItemBegin((VerilogCaseGenerateItem) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onCaseGenerateItemEnd((VerilogCaseGenerateItem) node);
      }
    });

    handlers.put(VerilogNode.Tag.CASE_STATEMENT, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onCaseStatementBegin((VerilogCaseStatement) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onCaseStatementEnd((VerilogCaseStatement) node);
      }
    });

    handlers.put(VerilogNode.Tag.CASE_STATEMENT_ITEM, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onCaseStatementItemBegin((VerilogCaseStatementItem) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onCaseStatementItemEnd((VerilogCaseStatementItem) node);
      }
    });

    handlers.put(VerilogNode.Tag.CODE, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onCodeBegin((VerilogCode) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onCodeEnd((VerilogCode) node);
      }
    });

    handlers.put(VerilogNode.Tag.DECLARATION, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onDeclarationBegin((VerilogDeclaration) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onDeclarationEnd((VerilogDeclaration) node);
      }
    });

    handlers.put(VerilogNode.Tag.DEFINE_PARAMETER, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onDefineParameterBegin((VerilogDefineParameter) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onDefineParameterEnd((VerilogDefineParameter) node);
      }
    });

    handlers.put(VerilogNode.Tag.DELAYED_STATEMENT, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onDelayedStatementBegin((VerilogDelayedStatement) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onDelayedStatementEnd((VerilogDelayedStatement) node);
      }
    });

    handlers.put(VerilogNode.Tag.DISABLE_STATEMENT, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onDisableStatementBegin((VerilogDisableStatement) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onDisableStatementEnd((VerilogDisableStatement) node);
      }
    });

    handlers.put(VerilogNode.Tag.GENERATE, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onGenerateBegin((VerilogGenerate) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onGenerateEnd((VerilogGenerate) node);
      }
    });

    handlers.put(VerilogNode.Tag.IF_GENERATE, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onIfGenerateBegin((VerilogIfGenerate) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onIfGenerateEnd((VerilogIfGenerate) node);
      }
    });

    handlers.put(VerilogNode.Tag.IF_GENERATE_BRANCH, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onIfGenerateBranchBegin((VerilogIfGenerateBranch) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onIfGenerateBranchEnd((VerilogIfGenerateBranch) node);
      }
    });

    handlers.put(VerilogNode.Tag.IF_STATEMENT, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onIfStatementBegin((VerilogIfStatement) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onIfStatementEnd((VerilogIfStatement) node);
      }
    });

    handlers.put(VerilogNode.Tag.IF_STATEMENT_BRANCH, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onIfStatementBranchBegin((VerilogIfStatementBranch) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onIfStatementBranchEnd((VerilogIfStatementBranch) node);
      }
    });

    handlers.put(VerilogNode.Tag.INSTANTIATION, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onInstantiationBegin((VerilogInstantiation) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onInstantiationEnd((VerilogInstantiation) node);
      }
    });

    handlers.put(VerilogNode.Tag.LOOP_GENERATE, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onLoopGenerateBegin((VerilogLoopGenerate) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onLoopGenerateEnd((VerilogLoopGenerate) node);
      }
    });

    handlers.put(VerilogNode.Tag.LOOP_STATEMENT, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onLoopStatementBegin((VerilogLoopStatement) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onLoopStatementEnd((VerilogLoopStatement) node);
      }
    });

    handlers.put(VerilogNode.Tag.MODULE, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onModuleBegin((VerilogModule) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onModuleEnd((VerilogModule) node);
      }
    });

    handlers.put(VerilogNode.Tag.NULL_STATEMENT, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onNullStatementBegin((VerilogNullStatement) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onNullStatementEnd((VerilogNullStatement) node);
      }
    });

    handlers.put(VerilogNode.Tag.PATH_DECLARATION, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onPathDeclarationBegin((VerilogPathDeclaration) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onPathDeclarationEnd((VerilogPathDeclaration) node);
      }
    });

    handlers.put(VerilogNode.Tag.PULSE_STYLE, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onPulseStyleBegin((VerilogPulseStyle) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onPulseStyleEnd((VerilogPulseStyle) node);
      }
    });

    handlers.put(VerilogNode.Tag.PORT, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onPortBegin((VerilogPort) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onPortEnd((VerilogPort) node);
      }
    });

    handlers.put(VerilogNode.Tag.PORT_CONNECTION, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onPortConnectionBegin((VerilogPortConnection) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onPortConnectionEnd((VerilogPortConnection) node);
      }
    });

    handlers.put(VerilogNode.Tag.PROCEDURE, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onProcedureBegin((VerilogProcedure) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onProcedureEnd((VerilogProcedure) node);
      }
    });

    handlers.put(VerilogNode.Tag.SHOW_CANCELLED, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onShowCancelledBegin((VerilogShowCancelled) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onShowCancelledEnd((VerilogShowCancelled) node);
      }
    });

    handlers.put(VerilogNode.Tag.SPECIFY, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onSpecifyBegin((VerilogSpecify) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onSpecifyEnd((VerilogSpecify) node);
      }
    });

    handlers.put(VerilogNode.Tag.TABLE, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onTableBegin((VerilogTable) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onTableEnd((VerilogTable) node);
      }
    });

    handlers.put(VerilogNode.Tag.TABLE_ENTRY, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onTableEntryBegin((VerilogTableEntry) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onTableEntryEnd((VerilogTableEntry) node);
      }
    });

    handlers.put(VerilogNode.Tag.TASK_STATEMENT, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onTaskStatementBegin((VerilogTaskStatement) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onTaskStatementEnd((VerilogTaskStatement) node);
      }
    });

    handlers.put(VerilogNode.Tag.TRIGGER_STATEMENT, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onTriggerStatementBegin((VerilogTriggerStatement) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onTriggerStatementEnd((VerilogTriggerStatement) node);
      }
    });

    handlers.put(VerilogNode.Tag.WAIT_STATEMENT, new VerilogNodeHandler() {
      @Override
      public Result onBegin(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onWaitStatementBegin((VerilogWaitStatement) node);
      }

      @Override
      public Result onEnd(final VerilogNodeVisitor visitor, final VerilogNode node) {
        return visitor.onWaitStatementEnd((VerilogWaitStatement) node);
      }
    });
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public final Result onBegin(final AbstractNode<VerilogNode.Tag> node) {
    return handlers.get(node.getTag()).onBegin(this, (VerilogNode) node);
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public final Result onEnd(final AbstractNode<VerilogNode.Tag> node) {
    return handlers.get(node.getTag()).onEnd(this, (VerilogNode) node);
  }
}
