`timescale 1ns/1ps

`include "ram.v"

// FIFO queue consisting of four 8-bit items
module fifo(clk, rst, val_rd, val_wr, data_in, val_out, data_out, is_full, is_ready);
    input         clk;
    input         rst;
    input         val_rd;
    input         val_wr;
    input  [31:0] data_in;
    output        val_out;
    output [31:0] data_out;
    output        is_full;
    output        is_ready;

    // RAM[i] is available
    reg val0, val1, val2, val3;

    reg  [1:0]    addr0;
    reg  [1:0]    addr1;
    reg  [1:0]    addr2;
    reg  [1:0]    addr3;

    reg  [1:0]    last;

    wire [1:0]    free_addr;
    wire [1:0]    addr_in;

    assign is_full = val0 & val1 & val2 & val3;

    assign free_addr = (~val0 ? 2'h0 : (~val1 ? 2'h1 : (~val2 ? 2'h2 : (~val3 ? 2'h3 : 2'hX))));

    assign addr_in = val_wr ? free_addr : (val_rd ? addr0 : 2'h0);

    ram memory
    (
        .clk(clk),
        .rst(rst),
        .val_rd(val_rd),
        .val_wr(val_wr),
        .addr_in(addr_in),
        .data_in(data_in),
        .val_out(val_out),
        .data_out(data_out),
        .is_ready(is_ready)
    );

    always @(posedge rst)
      begin
        val0  <= 1'b0;
        val1  <= 1'b0;
        val2  <= 1'b0;
        val3  <= 1'b0;
        addr0 <= 2'hX;
        addr1 <= 2'hX;
        addr2 <= 2'hX;
        addr3 <= 2'hX;
        last  <= 2'h0;
      end

    always @(posedge clk)
      begin
        if(~rst & val_wr)
          begin
            @(val_out);

            // Demultiplexor 1-to-4
            casex(last)
            2'h0: begin addr0 <= free_addr; val0 <= 1'b1; last <= 2'h1; end
            2'h1: begin addr1 <= free_addr; val1 <= 1'b1; last <= 2'h2; end
            2'h2: begin addr2 <= free_addr; val2 <= 1'b1; last <= 2'h3; end
            2'h3: begin addr3 <= free_addr; val3 <= 1'b1; last <= 2'h3; end
            default:;
            endcase
          end
        else if(~rst & val_rd)
          begin
            case(addr0)
            2'h0: val0 <= 1'b0;
            2'h1: val1 <= 1'b0;
            2'h2: val2 <= 1'b0;
            2'h3: val3 <= 1'b0;
            default:;
            endcase

            case(last)
            2'h0: last <= 2'h0;
            2'h1: last <= 2'h0;
            2'h2: last <= 2'h1;
            2'h3: last <= 2'h2;
            default:;
            endcase

            @(val_out);

            addr0 <= addr1;
            addr1 <= addr2;
            addr2 <= addr3;
            addr3 <= 2'hX;
          end
      end
endmodule /* fifo */

